#include "ICellMLToRDFTree.h"
#include <nsIRDFLiteral.h>
#include <nsCOMPtr.h>
#include <nsIRDFService.h>
#include <nsIRDFContainerUtils.h>
#include <nsClassHashtable.h>
#include <nsDataHashtable.h>
#include <nsInterfaceHashtable.h>

class CellMLToRDFTree;
typedef void (*rdf_populate_t)(CellMLToRDFTree*, nsISupports*,
                               nsIRDFContainer*);

class CellMLToRDFTree
  : public ICellMLToRDFTree
{
public:
  CellMLToRDFTree();
  ~CellMLToRDFTree();

  NS_DECL_ISUPPORTS;
  NS_DECL_ICELLMLTORDFTREE;

private:
  class ResourceAssociation
  {
  public:
    ResourceAssociation(const char* aType,
                        nsISupports* aElement,
                        CellMLToRDFTree* aCellMLToRDFTree,
                        nsIRDFContainer* aSeq,
                        rdf_populate_t aPopulate);

    const char* mType;
    nsCOMPtr<nsISupports> mElement;
    nsCOMPtr<nsIRDFContainer> mSeq;
    CellMLToRDFTree* mCellMLToRDFTree;
    rdf_populate_t mPopulate;
  };

public:
  nsCOMPtr<nsIRDFDataSource> mDatasource;
  nsClassHashtable<nsVoidPtrHashKey,ResourceAssociation> mHash;
  nsDataHashtable<nsCStringHashKey,int> mOpenSet;
  nsInterfaceHashtable<nsCStringHashKey, nsIRDFResource> mResourceHash;
  static void EnsureLiterals();
  static bool mSetupStatic;
  static nsIRDFResource * mRootResource, * mIvRootResource, * mTypeResource,
    * mPropResource, * mValueResource, * mValuePropResource, * mOpenResource,
    * mPlaceholderResource;
  static nsIRDFLiteral * mTrueLiteral, * mNumberLiteral, * mModelLiteral,
    * mComponentsLiteral, * mGroupLiteral, * mGroupsLiteral, * mConnectionsLiteral,
    * mConnectionLiteral, * mUnitsLiteral, * mImportsLiteral, * mVariablesLiteral,
    * mRolesLiteral, * mPubIntLiteral, * mPrivIntLiteral, * mBaseLiteral,
    * mDerivedLiteral, * mPrefixLiteral, * mExponentLiteral, * mOffsetLiteral,
    * mMultiplierLiteral, * mUnitLiteral, * mReversibleLiteral,
    * mIrreversibleLiteral, * mReactionLiteral, * mVariableRefLiteral,
    * mReactantLiteral, * mProductLiteral, * mRateLiteral, * mCatalystLiteral,
    * mActivatorLiteral, * mInhibitorLiteral, * mModifierLiteral,
    * mDirectionLiteral, * mForwardLiteral, * mReverseLiteral, * mBothLiteral,
    * mStoichiometryLiteral, * mDeltaVariableLiteral,
    * mRelationshipRefLiteral, * mNameLiteral, * mNamespaceLiteral,
    * mComponentRefLiteral, * mComponent1Literal, * mComponent2Literal,
    * mVariable1Literal, * mVariable2Literal, * mVariableMappingLiteral,
    * mMapVariablesLiteral, * mMathLiteral, * mEquationLiteral;
  static nsIRDFLiteral * mInterfaceIdLiteral[3];
  static nsCOMPtr<nsIRDFService> mRDFService;
  static nsCOMPtr<nsIRDFContainerUtils> mContUtils;
  static nsCOMPtr<nsIRDFDataSource> mPersistSource;

  void ExpandResourceInTree(nsIRDFResource* aResource);
  already_AddRefed<nsIRDFResource> GetAssociatedResource
    (const char* aType, nsISupports* aElement,
     rdf_populate_t aPopulate, PRBool aAlwaysOpen=PR_FALSE);
  void PopulateModelIntoSequences
  (
   cellml_apiIModel* aModel,
   nsIRDFContainer* aRootseq
  );
  void PopulateComponentsIntoSequences
  (
   cellml_apiICellMLComponentIterator* aCompIt,
   nsIRDFContainer* aCompSeq
  );
  void PopulateGroupsIntoSequences
  (
   cellml_apiIGroupIterator* aGroupIt,
   nsIRDFContainer* aGroupSeq
  );
  void PopulateConnectionsIntoSequences
  (
   cellml_apiIConnectionIterator* aConnIt,
   nsIRDFContainer* aConnSeq
  );
  void PopulateUnitsIntoSequences
  (
   cellml_apiIUnitsIterator* aUnitsIt,
   nsIRDFContainer* aUnitsSeq
  );
  void PopulateImportsIntoSequences
  (
   cellml_apiICellMLImportIterator* aCompIt,
   nsIRDFContainer* aImpSeq
  );
  void PopulateVariablesIntoSequences
  (
   cellml_apiICellMLVariableIterator* aVarIt,
   nsIRDFContainer* aVarSeq,
   PRBool aIgnoreNonIV = PR_FALSE
  );
  void PopulateReactionsIntoSequences
  (
   cellml_apiIReactionIterator* aReactIt,
   nsIRDFContainer* aReactSeq
  );
  void PopulateRelationshipRefsIntoSequences
  (
   cellml_apiIRelationshipRefIterator* aRRIt,
   nsIRDFContainer* aRRSeq
  );
  void PopulateComponentRefsIntoSequences
  (
   cellml_apiIComponentRefIterator* aCRIt,
   nsIRDFContainer* aCRSeq
  );
  void PopulateMapVariablesIntoSequences
  (
   cellml_apiIMapVariablesIterator* aMVIt,
   nsIRDFContainer* aMVSeq
  );
  void PopulateUnitIntoSequences
  (
   cellml_apiIUnitIterator* aUnIt,
   nsIRDFContainer* aUnitSeq
  );
  void PopulateImportComponentsIntoSequences
  (
   cellml_apiIImportComponentIterator* aMVIt,
   nsIRDFContainer* aICSeq
  );
  void PopulateImportUnitsIntoSequences
  (
   cellml_apiIImportUnitsIterator* aIUIt,
   nsIRDFContainer* aIUSeq
  );
  void PopulateVariableRefsIntoSequences
  (
   cellml_apiIVariableRefIterator* aVRIt,
   nsIRDFContainer* aVRSeq
  );
  void PopulateRolesIntoSequences
  (
   cellml_apiIRoleIterator* aRoleIt,
   nsIRDFContainer* aRoleSeq
  );
  void PopulateMathsIntoSequences
  (
   cellml_apiIMathMLElementIterator* aMEI,
   nsIRDFContainer* aMESeq
  );
  void PopulateEquationsIntoSequences
  (
   mathml_domIMathMLContainer* aEqContainer,
   PRUint32 aCount,
   nsIRDFContainer* aSeq
  );

  static PLDHashOperator rememberExpansionCallback(const void*,ResourceAssociation*,void*);
  void RememberExpansion(const nsIRDFResource*,ResourceAssociation*);
};
