#include <nsStringAPI.h>
#include "MathMLInputServices.h"
#include <nsCOMPtr.h>
#include "TextRepScanClass.h"

#include "TextRepLexer.h"

int textrep_parse(domIDocument* aDoc,
                  TextRepFlexLexer* aParser,
                  mathml_domIMathMLContentElement** aOutput);

MathMLInputServices::MathMLInputServices()
{
}

MathMLInputServices::~MathMLInputServices()
{
}


#define MATHML_NS NS_LITERAL_STRING("http://www.w3.org/1998/Math/MathML")
#define CELLML_1_1_NS NS_LITERAL_STRING("http://www.cellml.org/cellml/1.1#")
#define CELLML_1_0_NS NS_LITERAL_STRING("http://www.cellml.org/cellml/1.0#")

#define FLAG_TOPLEVEL 1
#define FLAG_BRACKETEQUAL 2

static void
AppendParameter(nsACString& params, const char* name, nsACString& value)
{
  if (params == "")
    params += "{";
  else
    params += " ";

  params += name;
  params += "=\"";
  const char* p = value.BeginReading();
  char c;
  while ((c = *p) != 0)
  {
    p++;
    if (c == '"' || c == '\\')
    {
      params += '\\';
      params += c;
    }
    else if (c == '\r')
      params += "\\r";
    else if (c == '\n')
      params += "\\n";
    else if (c == '\t')
      params += "\\t";
    else if (c == '\f')
      params += "\\f";
    else
      params += c;
  }
  params += "\"";
}

static void
AppendParameter(nsACString& params, const char* name, nsAString& value)
{
  NS_ConvertUTF16toUTF8 value8(value);
  AppendParameter(params, name, value8);
}

enum Precedence
{
  UNAMBIG,
  OR,
  AND,
  EQUALITY,
  PLUS_MINUS,
  TIMES_DIVIDE,
  NOT_UMINUS
};

#define CHILD_BVAR      0x0001
#define CHILD_LOWLIMIT  0x0002
#define CHILD_UPLIMIT   0x0004
#define CHILD_DEGREE    0x0008
#define CHILD_LOGBASE   0x0010

#define ATTR_BASE       0x0001
#define ATTR_CLOSURE    0x0002
#define ATTR_DEFURLONLY 0x0004
#define ATTR_ENCODING   0x0008
#define ATTR_ORDER      0x0010
#define ATTR_TYPE       0x0020

/* DEFURL implies encoding too */
#define ATTR_DEFURL     ATTR_DEFURLONLY | ATTR_ENCODING

static struct AttrDef {
  int mask;
  const char* attrName;
} AttrDefs[] =
{
  {ATTR_BASE, "base"},
  {ATTR_CLOSURE, "closure"},
  {ATTR_DEFURLONLY, "definitionURL"},
  {ATTR_ENCODING, "encoding"},
  {ATTR_ORDER, "order"},
  {ATTR_TYPE, "type"}
};

static struct SymbolInfo {
  const char* symName;
  enum {
    FUNCTIONAL,
    TOKEN,
    VARIABLE,
    CONSTANT,
    APPLY,
    PIECEWISE
  } Form;
  int Attrs;
  const char* text;
} SymbolInformation[] =
{
  {"apply", SymbolInfo::APPLY, 0, ""},
  {"bvar", SymbolInfo::FUNCTIONAL, 0, "bvar"},
  {"ci", SymbolInfo::VARIABLE, 0, ""},
  {"cn", SymbolInfo::CONSTANT, 0, ""},
  {"complexes", SymbolInfo::TOKEN, ATTR_DEFURL, "complexes"},
  {"condition", SymbolInfo::FUNCTIONAL, 0, "condition"},
  {"degree", SymbolInfo::FUNCTIONAL, 0, "degree"},
  {"domainofapplication", SymbolInfo::FUNCTIONAL, 0, "domainofapplication"},
  {"emptyset", SymbolInfo::TOKEN, ATTR_DEFURL, "emptyset"},
  {"eulergamma", SymbolInfo::TOKEN, ATTR_DEFURL, "eulergamma"},
  {"exponentiale", SymbolInfo::TOKEN, ATTR_DEFURL, "exponentiale"},
  {"false", SymbolInfo::TOKEN, ATTR_DEFURL, "false"},
  {"fn", SymbolInfo::FUNCTIONAL, 0, "function"},
  {"imaginaryi", SymbolInfo::TOKEN, ATTR_DEFURL, "imaginaryi"},
  {"infinity", SymbolInfo::TOKEN, ATTR_DEFURL, "infinity"},
  {"integers", SymbolInfo::TOKEN, ATTR_DEFURL, "integers"},
  {"interval", SymbolInfo::FUNCTIONAL, ATTR_CLOSURE, "interval"},
  {"lambda", SymbolInfo::FUNCTIONAL, 0, "lambda"},
  {"list", SymbolInfo::FUNCTIONAL, ATTR_ORDER|ATTR_TYPE, "list"},
  {"logbase", SymbolInfo::FUNCTIONAL, 0, "logbase"},
  {"lowlimit", SymbolInfo::FUNCTIONAL, 0, "lowlimit"},
  {"matrix", SymbolInfo::FUNCTIONAL, 0, "matrix"},
  {"matrixrow", SymbolInfo::FUNCTIONAL, 0, "row"},
  {"momentabout", SymbolInfo::FUNCTIONAL, 0, "momentabout"},
  {"naturalnumbers", SymbolInfo::TOKEN, ATTR_DEFURL, "naturalnumbers"},
  {"notanumber", SymbolInfo::TOKEN, ATTR_DEFURL, "notanumber"},
  {"pi", SymbolInfo::TOKEN, ATTR_DEFURL, "pi"},
  {"piecewise", SymbolInfo::PIECEWISE, 0, ""},
  {"primes", SymbolInfo::TOKEN, ATTR_DEFURL, "primes"},
  {"rationals", SymbolInfo::TOKEN, ATTR_DEFURL, "rationals"},
  {"reals", SymbolInfo::TOKEN, ATTR_DEFURL, "reals"},
  {"set", SymbolInfo::FUNCTIONAL, ATTR_ORDER|ATTR_TYPE, "set"},
  {"true", SymbolInfo::TOKEN, ATTR_DEFURL, "true"},
  {"uplimit", SymbolInfo::FUNCTIONAL, 0, "uplimit"},
  {"vector", SymbolInfo::FUNCTIONAL, 0, "vector"}
};

static struct OpInfo {
  const char* opName;
  enum {
    PREORDER,
    INORDER,
    DIFFSPLIT,
    PREORDER_OPTIONAL_BRACKET,
    INORDER_OR_PREORDER
  } DisplayAs;
  int SpecialForm;
  const char* text;
  Precedence prec;
} OperatorInformation[] = 
{
  {"and", OpInfo::INORDER, 0, "&", AND},
  {"diff", OpInfo::DIFFSPLIT, CHILD_BVAR | CHILD_DEGREE, "d", UNAMBIG},
  {"divide", OpInfo::INORDER, 0, "/", TIMES_DIVIDE},
  {"eq", OpInfo::INORDER, 0, "==", EQUALITY},
  {"geq", OpInfo::INORDER, 0, ">=", EQUALITY},
  {"gt", OpInfo::INORDER, 0, ">", EQUALITY},
  {"int", OpInfo::PREORDER, CHILD_BVAR | CHILD_LOWLIMIT | CHILD_UPLIMIT,
   "definite_integral", UNAMBIG},
  {"leq", OpInfo::INORDER, 0, "<=", EQUALITY},
  {"log", OpInfo::PREORDER, CHILD_LOGBASE, "log", UNAMBIG},
  {"lt", OpInfo::INORDER, 0, "<", EQUALITY},
  {"minus", OpInfo::INORDER_OR_PREORDER, 0, "-", PLUS_MINUS},
  {"neq", OpInfo::INORDER, 0, "!=", EQUALITY},
  {"not", OpInfo::PREORDER_OPTIONAL_BRACKET, 0, "!", NOT_UMINUS},
  {"or", OpInfo::INORDER, 0, "|", OR},
  {"partialdiff", OpInfo::DIFFSPLIT, CHILD_BVAR, "del", UNAMBIG},
  {"plus", OpInfo::INORDER, 0, "+", PLUS_MINUS},
  {"times", OpInfo::INORDER, 0, "*", TIMES_DIVIDE}
};

#define ATTR_TABLE_SIZE (sizeof(AttrDefs)/sizeof(AttrDefs[0]))
#define SYMBOL_TABLE_SIZE (sizeof(SymbolInformation)/sizeof(SymbolInformation[0]))
#define OPERATOR_TABLE_SIZE (sizeof(OperatorInformation)/sizeof(OperatorInformation[0]))

static OpInfo*
GetOperatorInformation(const nsACString& name)
{
  const char* sn = name.BeginReading();
  PRUint32 upper = OPERATOR_TABLE_SIZE - 1, lower = 0;
  OpInfo* ret = NULL;
  do
  {
    PRUint32 mid = (upper + lower) / 2;
    int cmp = strcmp(OperatorInformation[mid].opName, sn);
    if (cmp == 0)
    {
      ret = OperatorInformation + mid;
      break;
    }
    else if (upper == lower)
      break;
    else if (cmp < 0)
      lower = mid + 1;
    else
    {
      if (mid == lower)
        break;
      upper = mid - 1;
    }
  }
  while (true);
  return ret;
}

static SymbolInfo*
GetSymbolInformation(const nsACString& name)
{
  const char* sn = name.BeginReading();
  PRUint32 upper = SYMBOL_TABLE_SIZE - 1, lower = 0;
  SymbolInfo* ret = NULL;
  do
  {
    PRUint32 mid = (upper + lower) / 2;
    int cmp = strcmp(SymbolInformation[mid].symName, sn);
    if (cmp == 0)
    {
      ret = SymbolInformation + mid;
      break;
    }
    else if (upper == lower)
      break;
    else if (cmp < 0)
      lower = mid + 1;
    else
    {
      if (mid == lower)
        break;
      upper = mid - 1;
    }
  }
  while (true);
  return ret;
}

static void RecursivelyAppendCode(mathml_domIMathMLContentElement* aMathML,
                                  nsACString& aOutput, int flags,
                                  Precedence contextPrec);

static void
RecursivelyAppendCodePiecewise(mathml_domIMathMLPiecewiseElement* aMathML,
                               nsACString& aOutput, nsACString& aParams)
{
  aOutput += "piecewise";
  aOutput += aParams;
  aOutput += "(";
  PRUint32 i = 0;
  while (PR_TRUE)
  {
    i++;
    nsCOMPtr<mathml_domIMathMLCaseElement> ca;
    aMathML->GetCase(i, getter_AddRefs(ca));
    if (ca == nsnull)
      break;

    nsCString params;
    nsString v;
    ca->GetClassName(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "class", v);
    ca->GetMathElementStyle(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "style", v);
    ca->GetId(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "id", v);
    ca->GetXref(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "xref", v);
    ca->GetHref(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "href", v);
    if (!params.Equals(""))
      params += "}";

    if (i != 1)
      aOutput += " ";

    aOutput += "case";
    aOutput += params;
    aOutput += " ";
    nsCOMPtr<mathml_domIMathMLContentElement> ce;
    ca->GetCaseCondition(getter_AddRefs(ce));
    RecursivelyAppendCode(ce, aOutput, 0, UNAMBIG);
    aOutput += " then ";
    ca->GetCaseValue(getter_AddRefs(ce));
    RecursivelyAppendCode(ce, aOutput, 0, UNAMBIG);
  }
  nsCOMPtr<mathml_domIMathMLContentElement> ce;
  aMathML->GetOtherwise(getter_AddRefs(ce));
  if (ce != nsnull)
  {
    if (i != 1)
      aOutput += " ";
    aOutput += "else";

    nsCString params;
    nsString v;
    ce->GetClassName(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "class", v);
    ce->GetMathElementStyle(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "style", v);
    ce->GetId(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "id", v);
    ce->GetXref(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "xref", v);
    ce->GetHref(v);
    if (!v.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "href", v);
    if (!params.Equals(""))
      params += "}";

    aOutput += params;
    aOutput += " ";
    RecursivelyAppendCode(ce, aOutput, 0, UNAMBIG);
  }
  aOutput += ")";
}

static void
RecursivelyAppendCodeApply(mathml_domIMathMLApplyElement* aMathML,
                           nsACString& aOutput,
                           int flags,
                           Precedence contextPrec,
                           nsACString& aParams)
{
  nsCOMPtr<mathml_domIMathMLElement> op;
  aMathML->GetOperator(getter_AddRefs(op));
  nsString nsURI;

  if (op == nsnull)
  {
    aOutput.Append("<no operator yet>");
    return;
  }

  PRBool considerSpecial = PR_TRUE;
  op->GetNamespaceURI(nsURI);
  if (!nsURI.Equals(MATHML_NS))
  {
    considerSpecial = PR_FALSE;
    AppendParameter(aParams, "namespace", nsURI);
  }

  if (!aParams.Equals(""))
    aParams += "}";

  nsString lname;
  op->GetLocalName(lname);
  NS_ConvertUTF16toUTF8 lname8(lname);

  OpInfo* oi = nsnull;
  OpInfo oiDef;
  if (considerSpecial)
    oi = GetOperatorInformation(lname8);

  nsCOMPtr<mathml_domIMathMLContentElement>
    bvar, lowlimit, uplimit, degree, logbase;
  int count = 0;

  if (oi == nsnull)
  {
    oi = &oiDef;
    oiDef.DisplayAs = OpInfo::PREORDER;
    oiDef.SpecialForm = 0;
    oiDef.text = lname8.get();
    oiDef.prec = UNAMBIG;
  }
  else if (oi->SpecialForm != 0 || oi->DisplayAs == OpInfo::INORDER_OR_PREORDER)
  {
    nsCOMPtr<domINodeList> nl;
    aMathML->GetChildNodes(getter_AddRefs(nl));
    PRUint32 i = 0, l;
    nl->GetLength(&l);
    for (i = 0; i < l; i++)
    {
      nsCOMPtr<domINode> n;
      nl->Item(i, getter_AddRefs(n));
      nsCOMPtr<mathml_domIMathMLContentElement> el(do_QueryInterface(n));
      if (el == nsnull)
        continue;
      nsString ln2;
      el->GetLocalName(ln2);
#define SPECIAL_QUAL(x,X) \
      if ((oi->SpecialForm & CHILD_##X) && \
          ln2.Equals(NS_LITERAL_STRING(#x))) \
      { \
        nsCOMPtr<mathml_domIMathMLContainer> cc(do_QueryInterface(el)); \
        nsCOMPtr<mathml_domIMathMLElement> tel; \
        cc->GetArgument(1, getter_AddRefs(tel)); \
        x = do_QueryInterface(tel); \
      }
      SPECIAL_QUAL(bvar, BVAR);
      SPECIAL_QUAL(lowlimit, LOWLIMIT);
      SPECIAL_QUAL(uplimit, UPLIMIT);
      SPECIAL_QUAL(degree, DEGREE);
      SPECIAL_QUAL(logbase, LOGBASE);
      count++;
    }
  }

  PRBool isPreorder = PR_FALSE;
  Precedence innerPrec = oi->prec;

  if (oi->DisplayAs == OpInfo::PREORDER ||
      oi->DisplayAs == OpInfo::DIFFSPLIT ||
      oi->DisplayAs == OpInfo::PREORDER_OPTIONAL_BRACKET)
    isPreorder = PR_TRUE;

  // INORDER_OR_PREORDER (e.g. minus) get their precedence changed
  // if they have only one child (pre-order case)...
  if (oi->DisplayAs == OpInfo::INORDER_OR_PREORDER && count <= 2)
  {
    innerPrec = NOT_UMINUS;
    isPreorder = PR_TRUE;
  }

  if ((innerPrec < contextPrec && innerPrec != UNAMBIG) ||
      ((flags & FLAG_BRACKETEQUAL) && innerPrec == contextPrec))
    aOutput += "(";

  if (isPreorder)
  {
    aOutput += oi->text;
    aOutput += aParams;
    if (oi->DisplayAs != OpInfo::PREORDER_OPTIONAL_BRACKET)
      aOutput += "(";
  }

  // Now we have to iterate through the arguments...
  PRBool opHit = PR_TRUE, firstHit = PR_TRUE, secondHit = PR_TRUE;
  nsCOMPtr<domINodeList> nl;
  aMathML->GetChildNodes(getter_AddRefs(nl));
  PRUint32 i, l;
  nl->GetLength(&l);
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<domINode> n;
    nl->Item(i, getter_AddRefs(n));
    nsCOMPtr<mathml_domIMathMLContentElement> el(do_QueryInterface(n));
    if (el == nsnull)
      continue;
    if (opHit)
    {
      opHit = PR_FALSE;
      continue;
    }
    nsString ln2;
    el->GetNamespaceURI(ln2);
    if (!ln2.Equals(MATHML_NS))
      continue;
    el->GetLocalName(ln2);
    if ((oi->SpecialForm & CHILD_BVAR) &&
        ln2.Equals(NS_LITERAL_STRING("bvar")))
      continue;
    else if ((oi->SpecialForm & CHILD_LOWLIMIT) &&
             ln2.Equals(NS_LITERAL_STRING("lowlimit")))
      continue;
    else if ((oi->SpecialForm & CHILD_LOWLIMIT) &&
             ln2.Equals(NS_LITERAL_STRING("uplimit")))
      continue;
    else if ((oi->SpecialForm & CHILD_DEGREE) &&
             ln2.Equals(NS_LITERAL_STRING("degree")))
      continue;
    else if ((oi->SpecialForm & CHILD_LOGBASE) &&
             ln2.Equals(NS_LITERAL_STRING("logbase")))
      continue;
    
    // We have an argument. See if it is first...
    if (!firstHit)
    {
      if (isPreorder)
        aOutput += ", ";
      else
      {
        aOutput += " ";
        // == is written as = if at the toplevel...
        if (!(flags & FLAG_TOPLEVEL) || strcmp(oi->text, "=="))
          aOutput += oi->text;
        else
          aOutput += "=";
        if (secondHit)
        {
          aOutput += aParams;
          secondHit = PR_FALSE;
        }
        aOutput += " ";
      }
    }

    // Now, put the code for the child node...
    RecursivelyAppendCode(el, aOutput, firstHit ? FLAG_BRACKETEQUAL : 0, innerPrec);

    if (firstHit)
      firstHit = PR_FALSE;
  }

  if (oi->DisplayAs == OpInfo::DIFFSPLIT)
  {
    firstHit = PR_TRUE;
    aOutput += ")/";
    aOutput += oi->text;
    aOutput += "(";
  }

  if (bvar != nsnull)
  {
    if (!firstHit)
      aOutput += ", ";

    RecursivelyAppendCode(bvar, aOutput, firstHit ? FLAG_BRACKETEQUAL : 0,
                          innerPrec);
    if (firstHit)
      firstHit = PR_FALSE;
  }
  if (lowlimit != nsnull)
  {
    if (!firstHit)
      aOutput += ", ";

    RecursivelyAppendCode(lowlimit, aOutput, firstHit ? FLAG_BRACKETEQUAL : 0,
                          innerPrec);
    if (firstHit)
      firstHit = PR_FALSE;
  }
  if (uplimit != nsnull)
  {
    if (!firstHit)
      aOutput += ", ";

    RecursivelyAppendCode(uplimit, aOutput, firstHit ? FLAG_BRACKETEQUAL : 0,
                          innerPrec);
    if (firstHit)
      firstHit = PR_FALSE;
  }
  if (degree != nsnull)
  {
    if (!firstHit)
      aOutput += ", ";

    RecursivelyAppendCode(degree, aOutput, firstHit ? FLAG_BRACKETEQUAL : 0,
                          innerPrec);
    if (firstHit)
      firstHit = PR_FALSE;
  }
  if (logbase != nsnull)
  {
    if (!firstHit)
      aOutput += ", ";

    RecursivelyAppendCode(logbase, aOutput, firstHit ? FLAG_BRACKETEQUAL : 0,
                          innerPrec);
    if (firstHit)
      firstHit = PR_FALSE;
  }

  if (isPreorder && oi->DisplayAs != OpInfo::PREORDER_OPTIONAL_BRACKET)
  {
    aOutput += ")";
  }

  if ((innerPrec < contextPrec && innerPrec != UNAMBIG) ||
      ((flags & FLAG_BRACKETEQUAL) && innerPrec == contextPrec))
    aOutput += ")";
}

static void
RecursivelyAppendCode
(
 mathml_domIMathMLContentElement* aMathML,
 nsACString& aOutput,
 int flags,
 Precedence contextPrec
)
{
  if (aMathML == nsnull)
  {
    aOutput += "[undefined]";
    return;
  }

  nsString ln;
  aMathML->GetLocalName(ln);

  nsCString params;
  nsString v;
  aMathML->GetClassName(v);
  if (!v.Equals(NS_LITERAL_STRING("")))
    AppendParameter(params, "class", v);
  aMathML->GetMathElementStyle(v);
  if (!v.Equals(NS_LITERAL_STRING("")))
    AppendParameter(params, "style", v);
  aMathML->GetId(v);
  if (!v.Equals(NS_LITERAL_STRING("")))
    AppendParameter(params, "id", v);
  aMathML->GetXref(v);
  if (!v.Equals(NS_LITERAL_STRING("")))
    AppendParameter(params, "xref", v);
  aMathML->GetHref(v);
  if (!v.Equals(NS_LITERAL_STRING("")))
    AppendParameter(params, "href", v);

  NS_ConvertUTF16toUTF8 ln8(ln);
  SymbolInfo* si = GetSymbolInformation(ln8);
  if (si == nsnull)
  {
    aOutput += "unsupported-element";
    return;
  }

  PRUint32 i;
  for (i = 0; i < ATTR_TABLE_SIZE; i++)
  {
    if (AttrDefs[i].mask && si->Attrs)
    {
      NS_ConvertUTF8toUTF16 att16(AttrDefs[i].attrName);
      nsString value;
      aMathML->GetAttribute(att16, value);
      AppendParameter(params, AttrDefs[i].attrName, value);
    }
  }

  if (si->Form == SymbolInfo::APPLY)
  {
    nsCOMPtr<mathml_domIMathMLApplyElement> mma(do_QueryInterface(aMathML));
    RecursivelyAppendCodeApply(mma, aOutput, flags, contextPrec, params);
    return;
  }
  else if (si->Form == SymbolInfo::CONSTANT)
  {
    nsString u;
    aMathML->GetAttributeNS(CELLML_1_1_NS,
                            NS_LITERAL_STRING("units"),
                            u);
    if (!u.Equals(NS_LITERAL_STRING("")))
      AppendParameter(params, "units", u);
    else
    {
      aMathML->GetAttributeNS(CELLML_1_0_NS,
                              NS_LITERAL_STRING("units"),
                              u);
      if (!u.Equals(NS_LITERAL_STRING("")))
        AppendParameter(params, "units", u);
    }

    if (!params.Equals(""))
      params += "}";

    nsCOMPtr<mathml_domIMathMLCnElement> c(do_QueryInterface(aMathML));
    nsCOMPtr<domINode> n;
    c->GetArgument(1, getter_AddRefs(n));
    if (n == nsnull)
      aOutput += "invalid-constant";
    nsCOMPtr<domIText> tn(do_QueryInterface(n));
    if (tn == nsnull)
      aOutput += "invalid-constant";
    nsString value;
    tn->GetData(value);
    NS_ConvertUTF16toUTF8 value8(value);
    aOutput += value8;
    aOutput += params;
    return;
  }

  if (!params.Equals(""))
    params += "}";

  if (si->Form == SymbolInfo::PIECEWISE)
  {
    nsCOMPtr<mathml_domIMathMLPiecewiseElement> mp(do_QueryInterface(aMathML));
    RecursivelyAppendCodePiecewise(mp, aOutput, params);
  }
  else if (si->Form == SymbolInfo::VARIABLE)
  {
    nsCOMPtr<mathml_domIMathMLContentToken> t(do_QueryInterface(aMathML));
    nsCOMPtr<domINode> n;
    t->GetArgument(1, getter_AddRefs(n));
    if (n == nsnull)
      aOutput += "invalid-var";
    nsCOMPtr<domIText> tn(do_QueryInterface(n));
    if (tn == nsnull)
      aOutput += "invalid-var";
    nsString value;
    tn->GetData(value);
    NS_ConvertUTF16toUTF8 value8(value);
    SymbolInfo* si2 = GetSymbolInformation(value8);
    if (si2 != nsnull)
      aOutput += "$";
    aOutput += value8;
    aOutput += params;
  }
  else if (si->Form == SymbolInfo::TOKEN)
  {
    aOutput += si->text;
    aOutput += params;
  }
  else if (si->Form == SymbolInfo::FUNCTIONAL)
  {
    aOutput += si->text;
    aOutput += params;
    aOutput += "(";
    PRBool firstHit = PR_TRUE;

    nsCOMPtr<domINodeList> nl;
    aMathML->GetChildNodes(getter_AddRefs(nl));
    PRUint32 l;
    nl->GetLength(&l);
    for (i = 0; i < l; i++)
    {
      nsCOMPtr<domINode> n;
      nl->Item(i, getter_AddRefs(n));
      nsCOMPtr<mathml_domIMathMLContentElement> el(do_QueryInterface(n));
      if (el == nsnull)
        continue;
      if (!firstHit)
        aOutput += ", ";
      else
        firstHit = PR_FALSE;

      RecursivelyAppendCode(el, aOutput, 0, UNAMBIG);
    }
    aOutput += ")";
  }
}

NS_IMPL_ISUPPORTS1(MathMLInputServices, IMathMLInputServices);

NS_IMETHODIMP
MathMLInputServices::MathMLToInputFormat
(
 mathml_domIMathMLApplyElement* aMathML,
 nsACString& aOutput
)
{
  NS_ENSURE_TRUE(aMathML, NS_ERROR_INVALID_ARG);

  aOutput.Assign("");
  RecursivelyAppendCode(aMathML, aOutput, FLAG_TOPLEVEL, UNAMBIG);
  return NS_OK;
}

NS_IMETHODIMP
MathMLInputServices::InputFormatToMathML
(
 domIDocument* aDoc,
 const nsACString& aHumanReadable,
 domIElement** aOutput
)
{
  NS_ENSURE_TRUE(aDoc, NS_ERROR_INVALID_ARG);

  TextRepLexer parser(aHumanReadable.BeginReading());
  parser.mColumn = 0;
  nsCOMPtr<mathml_domIMathMLContentElement> output;
  textrep_parse(aDoc, &parser, getter_AddRefs(output));
  if (output != nsnull && !parser.mEOF)
  {
    parser.LexerError("Garbage follows a valid expression.");
    output = nsnull;
  }

  nsCOMPtr<domIElement> el;
  if (output != nsnull)
  {
    *aOutput = NS_STATIC_CAST(domIElement*, output);
  }
  else
  {
    aDoc->CreateElementNS
    (
     NS_LITERAL_STRING("http://www.cellml.org/tools/mathml-input/error#"),
     NS_LITERAL_STRING("error"),
     getter_AddRefs(el)
    );
    nsCString& msg = parser.GetError();
    NS_ConvertUTF8toUTF16 msg16(msg);
    el->SetAttribute(NS_LITERAL_STRING("problem"),
                     msg16);
    char buf[30];
    sprintf(buf, "%u", parser.mColumn);
    NS_ConvertUTF8toUTF16 col16(buf);

    el->SetAttribute(NS_LITERAL_STRING("column"),
                     col16);
    *aOutput = el;
  }

  NS_ADDREF(*aOutput);

  return NS_OK;
}
